"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var globals_1 = require("@jest/globals");
var select_1 = require("./select");
(0, globals_1.describe)('parseSelect', function () {
    (0, globals_1.test)('basic fields', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT id, name FROM users;')).toEqual({
        command: 'SELECT',
        fields: [{ name: 'id' }, { name: 'name' }],
        tables: [{ name: 'users' }]
    }); });
    (0, globals_1.test)('field alias', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT id, name as username FROM users;')).toEqual({
        command: 'SELECT',
        fields: [{ name: 'id' }, { name: 'username', expression: 'name' }],
        tables: [{ name: 'users' }]
    }); });
    (0, globals_1.test)('wildcard', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT * FROM public.users;')).toEqual({
        command: 'SELECT',
        fields: [{ name: '*' }],
        tables: [{ name: 'public.users' }]
    }); });
    (0, globals_1.test)('table alias', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT u.id FROM public.users u;')).toEqual({
        command: 'SELECT',
        fields: [{ name: 'id', scope: 'u' }],
        tables: [{ name: 'u', alias: 'public.users' }]
    }); });
    (0, globals_1.test)('simple join', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT u.id, e.id, o.id FROM users u JOIN events e ON u.id = e.created_by JOIN organizations o ON o.id = e.organization_id;')).toEqual({
        command: 'SELECT',
        fields: [{ name: 'id', scope: 'u' }, { name: 'id', scope: 'e' }, { name: 'id', scope: 'o' }],
        tables: [
            { name: 'u', alias: 'users' },
            { name: 'e', alias: 'events', on: 'u.id = e.created_by' },
            { name: 'o', alias: 'organizations', on: 'o.id = e.organization_id' }
        ]
    }); });
    (0, globals_1.test)('join and wildcard', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT u.*, e.id FROM users u JOIN events e ON u.id = e.created_by;')).toEqual({
        command: 'SELECT',
        fields: [{ name: '*', scope: 'u' }, { name: 'id', scope: 'e' }],
        tables: [{ name: 'u', alias: 'users' }, { name: 'e', alias: 'events', on: 'u.id = e.created_by' }]
    }); });
    (0, globals_1.test)('join and anonymous fields', function () { return (0, globals_1.expect)((0, select_1.parseSelect)('SELECT slug, project_id FROM users u JOIN events e ON u.id = e.created_by;')).toEqual({
        command: 'SELECT',
        fields: [{ name: 'slug' }, { name: 'project_id' }],
        tables: [{ name: 'u', alias: 'users' }, { name: 'e', alias: 'events', on: 'u.id = e.created_by' }]
    }); });
    (0, globals_1.test)('pagination', function () { return (0, globals_1.expect)((0, select_1.parseSelect)("SELECT id, name FROM users WHERE provider='github' ORDER BY id LIMIT 5 OFFSET 10;")).toEqual({
        command: 'SELECT',
        fields: [{ name: 'id' }, { name: 'name' }],
        tables: [{ name: 'users' }],
        where: "provider='github'",
        sort: 'id',
        limit: 5,
        offset: 10
    }); });
    (0, globals_1.describe)('parseSelectField', function () {
        (0, globals_1.test)('field basic', function () { return (0, globals_1.expect)((0, select_1.parseSelectField)('id'))
            .toEqual({ name: 'id' }); });
        (0, globals_1.test)('field with alias', function () { return (0, globals_1.expect)((0, select_1.parseSelectField)('id as user_id'))
            .toEqual({ name: 'user_id', expression: 'id' }); });
        (0, globals_1.test)('field with scope', function () { return (0, globals_1.expect)((0, select_1.parseSelectField)('u.id'))
            .toEqual({ name: 'id', scope: 'u' }); });
        (0, globals_1.test)('field with scope and alias', function () { return (0, globals_1.expect)((0, select_1.parseSelectField)('u.id as user_id'))
            .toEqual({ name: 'user_id', expression: 'id', scope: 'u' }); });
        (0, globals_1.test)('expression field', function () { return (0, globals_1.expect)((0, select_1.parseSelectField)("count(distinct to_char(e.created_at, 'yyyy-mm-dd')) as active_days"))
            .toEqual({ name: 'active_days', expression: "count(distinct to_char(e.created_at, 'yyyy-mm-dd'))" }); });
    });
    (0, globals_1.describe)('parseSelectTable', function () {
        (0, globals_1.test)('table basic', function () { return (0, globals_1.expect)((0, select_1.parseSelectTable)('users'))
            .toEqual({ name: 'users' }); });
        (0, globals_1.test)('table with schema', function () { return (0, globals_1.expect)((0, select_1.parseSelectTable)('public.users'))
            .toEqual({ name: 'public.users' }); });
        (0, globals_1.test)('table with alias', function () { return (0, globals_1.expect)((0, select_1.parseSelectTable)('users u'))
            .toEqual({ name: 'u', alias: 'users' }); });
        (0, globals_1.test)('table with schema and alias', function () { return (0, globals_1.expect)((0, select_1.parseSelectTable)('public.users u'))
            .toEqual({ name: 'u', alias: 'public.users' }); });
    });
    (0, globals_1.describe)('parseSelectJoin', function () {
        (0, globals_1.test)('join basic', function () { return (0, globals_1.expect)((0, select_1.parseSelectJoin)('events ON users.id = events.created_by'))
            .toEqual({ name: 'events', on: 'users.id = events.created_by' }); });
        (0, globals_1.test)('join with alias', function () { return (0, globals_1.expect)((0, select_1.parseSelectJoin)('events e ON u.id = e.created_by'))
            .toEqual({ name: 'e', alias: 'events', on: 'u.id = e.created_by' }); });
    });
});
